/*
  SaveBin. 1.2
  (c) Alexey Makhmutov (aka Lerm) 2004
*/

#include <stdrom.h>
#include <define.h>
#include <libc.h>
#include <l_define.h>
#include <l_libc.h>

#include <ziface.h>
#include <dialogs.h>

#define BLOCK_SIZE 3072
#define SBLOCK  64*1024L
#define DBLOCK 128*1024L

typedef struct
{
  byte mMode;
  byte sMode;
  int start;
  unsigned long length;
  char name[16];
} AddInInfo;

typedef struct
{
  char addinsig[8];      /* = {0x00, 0xFF, 'C','A','S','I','O', 0x03}; */
  char model[4];         /* = {'Z','4','8','6'}; */
  char headerversion[4]; /* = {'0','1','0','0'}; */
  unsigned short status;
  unsigned short mode;
  char addinname[16];
  unsigned long addinlength;
  char compiledate[8];   /* e.g. "20020219" stands for Feb. 19th, 2002 */
  char compiletime[4];   /* e.g. "1246" is 46 minutes past 12 o'clock */
  char version[4];       /* "0120" stands for version 1.20 */
  char libdate[8];       /* e.g. "20000215" stands for Feb. 15th, 2000 */
  char libtime[4];       /* e.g. "0940" is 40 minutes past 9 o'clock */
  char libver[4];        /* e.g. "0100" stands for version 1.00 */
  unsigned long offsicon;
  unsigned long offslicon;
  char comment[64];
} AddInHeader;

AddInInfo addins[16];
int addCount=0;

char* nameList[]={ addins[0].name,  addins[1].name,  addins[2].name,  addins[3].name,
                   addins[4].name,  addins[5].name,  addins[6].name,  addins[7].name,
                   addins[8].name,  addins[9].name, addins[10].name, addins[11].name,
                  addins[12].name, addins[13].name, addins[14].name, addins[15].name};

/*
 * SwitchBank
 *
 * makes RAM or FLASH memory accessible at certain bank
 *
 * addr A31-A16 of source address (must be even)
 *      use 0000 - 00FE for RAM (only 0000 will work)
 *      use 0100 - 010E for program flash (1st MB, PVOS)
 *      use 0110 - 011E for program flash (2nd MB, Add-In)
 *      use 0120 - 013E for data flash (4MB data)
 *
 * bank destination bank
 *      use 0 for address space 0000:0000 (forbidden)
 *      use 1 for address space 2000:0000
 *      use 2 for address space 4000:0000
 *      use 3 for address space 6000:0000
 *      use 4 for address space 8000:0000 (forbidden)
 *      use 5 for address space A000:0000
 *      use 6 for address space C000:0000 (forbidden)
 *
 * remarks:
 *   you can not switch bank 0, 4 and 6, because that will crash your
 *   PV; bank 0 contains the interrupt vectors, and every program
 *   expects to find RAM here; bank 4 contains your Add-In, so
 *   switching this would make your Add-In vanishing; bank 6 contains
 *   the BIOS, your PV can not operate without it
 *
 * examples:
 *
 *   SwitchBank(0x0000,3);
 *     will make 128kB RAM accessible at address 6000:0000 so RAM can
 *     be accessed at 0000:0000 and 6000:0000 in the same way (which
 *     is possible but stupid, you still have only 128kB of RAM)
 *
 *   SwitchBank(0x0104,5);
 *     will make Fonts accessible at address A000:0000 and Graphics
 *     accessible at B000:0000
 *
 *   SwitchBank(0x0120,1);
 *     will make first 128kB of data flash accessible at 2000:0000
 *
 */
void _asm_switchbank(char *,int,int);
#define SwitchBank(addr,bank) _asm_switchbank("\n MOV BH,AL\n MOV AL,AH\n MOV AH,BL\n OR AH,80h\n INT 0C8h\n MOV AL,BH\n MOV AH,BL\n INT 0C8h\n",addr,bank);

FILE_BUF fb;
FILE_INFO fi;
byte mode,submode;
char fileName[16]="StoredAddIn";
int storeFlag=0;
int saveFlag=0;

/* Main form */
int onSave(Event*,Control*);
int onExit(Event*,Control*);
void mainHandler(Event*,Form*);

Label label1={"Installed AddIns",IB_PFONT2,ALIGN_CENTER};
ListBox listBox={0,0,0,nameList};
Button bSave={B_PRESS,FALSE,"Save"};
Button bExit={B_PRESS,FALSE,"Exit"};

Control cList1[]={
                  {C_LABEL,   0,  0, 160,  0, TRUE, TRUE, NULL, &label1},
                  {C_LIST,    0, 10, 160, 160-10-15-2, TRUE, TRUE, NULL, &listBox},
                  {C_BUTTON, 159-40*2-2, 159-15, 40, 15, TRUE, TRUE, (CHandler)onSave, &bSave},
                  {C_BUTTON, 159-40, 159-15, 40, 15, TRUE, TRUE, (CHandler)onExit, &bExit}
                 };

Form mainForm={0,0,160,160,FORM_INACTIVE,elements(cList1),cList1,(FHandler)mainHandler};

/* Input form */
#define INPUT_BASE 15

int onOk(Event*,Control*);
int onCancel(Event*,Control*);
void inputHandler(Event*,Form*);

Icon icon1={108};
Label label2={"Enter filename:",IB_PFONT2,ALIGN_CENTER};
EditBox edit1={15,fileName};
Button bOk={B_PRESS,FALSE,"Ok"};
Button bCancel={B_PRESS,FALSE,"Cancel"};

Control cList2[]={
                  {C_ICON,   16, INPUT_BASE,    128, 62, TRUE, TRUE, NULL, &icon1},
                  {C_LABEL,  20, INPUT_BASE+6,  120, 0,  TRUE, TRUE, NULL, &label2},
                  {C_EDIT,   22, INPUT_BASE+21, 116, 12, TRUE, TRUE, NULL, &edit1},
                  {C_BUTTON, 56, INPUT_BASE+41, 40,16,TRUE,TRUE, (CHandler)onOk, &bOk},
                  {C_BUTTON, 98, INPUT_BASE+41, 40,16,TRUE,TRUE, (CHandler)onCancel, &bCancel}
                 };

Form inputForm={16,INPUT_BASE,128,62,FORM_INACTIVE,elements(cList2),cList2,(FHandler)inputHandler};

/* Wait form */
#define WAIT_BASE 55

void onActivateForm(Event*,Form*);
Icon icon2={107};/*128x50*/
Icon icon3={24};
ProgressBar pBar={0,0};

Control cList3[]={
                  {C_ICON,     16, WAIT_BASE, 128, 50, TRUE, TRUE, NULL, &icon2},
                  {C_ICON,     68, WAIT_BASE+6, 24, 24, TRUE, TRUE, NULL, &icon3},
                  {C_PROGRESS, 22, WAIT_BASE+32, 116, 12, TRUE, TRUE, NULL, &pBar}
                 };
Form waitForm={16,WAIT_BASE,128,50,FORM_INACTIVE,elements(cList3),cList3,(FHandler)onActivateForm};

/* Func */

void insertString(char* buffer, char* pattern, char* part)
{
  int i=0,j=0,first=1;
  if (pattern[0])
  {
    do{
      char c=pattern[i];
      if (first&&c=='%')
      {
        if (part[0])
        {
          int k=0;
          do{
            buffer[j]=part[k];
            j++;
            k++;
          } while (part[k]);
        }
        first=0;
      }
      else
      {
        buffer[j]=c;
        j++;
      }
      i++;
    } while (pattern[i]);
  }
 
  buffer[j]=0;
}

int initFile(char* filename)
{/* Open file */
  if (!LibGetAllEntry(filename, &mode, &submode)
      && LibSubEntrySave(filename, &submode)
      && LibGetAllEntry(filename, &mode, &submode)) 
    return 1;
  else
    return 0;
}

void initBlock(int size)
{/* Create new record */
  fi.fp=0xFFFF;
  fi.kind=FILE_KIND_BIN;

  fb.fsb_main_entry_ = mode;
  fb.fsb_sub_entry_  = submode;
  fb.fsb_scrt_info_  = 0x80;
  fb.fbuf.bin.char_num = size;
}

int saveBlock(byte far* pointer, long size)
{/* Save memory block */
  int blockSize=BLOCK_SIZE;

  while(size>0)
  {
    if (size<blockSize)
      blockSize=size;
    initBlock(blockSize);
    far_memcpy(fb.fbuf.bin.bin_buf,pointer,blockSize);
    if (!LibFileWrite(&fb,&fi))
      return 0;
    pBar.current+=blockSize;
    drawControl(cList3+2);
    pointer+=blockSize;
    size-=blockSize;
  }
  return 1;
}

byte far* getBlock(int num)
{/* Get addin pointer */
  SwitchBank(0x110+(num&0xFE),5);
  if (num&1)
    return MK_FP(0xB000,0x0000);
  else
    return MK_FP(0xA000,0x0000);
}

int saveAddIn(int num)
{/* Save addin */
  if (num<addCount)
  {
    int start=addins[num].start;
    long len=addins[num].length;
    if (len>SBLOCK)
      return saveBlock(getBlock(start),SBLOCK)&&saveBlock(getBlock(start+1),len-SBLOCK);
    else
      return saveBlock(getBlock(start),len);
  }
  return 0;
}

void scanFlash()
{/* Detect all addins */
  int block=0,i;
  do
  {
    AddInHeader far* head=(AddInHeader far*)getBlock(block);
    if (head->addinsig[0]==0&&head->addinsig[1]!=0)
    {
      addins[addCount].mMode=(head->mode>>8)&0xFF;
      addins[addCount].sMode=(head->mode)&0xFF;
      addins[addCount].start=block;
      addins[addCount].length=head->addinlength;
      //far_strcpy(addins[addCount].name,head->addinname);
      addCount++;
      if (head->addinlength>SBLOCK)
      {
        block++;
      }
    }  
    block++;
  } while(block<16);
  for(i=0; i<addCount; i++)
  {
    byte mByte=addins[i].mMode;
    byte sByte=addins[i].sMode;
    char* name=addins[i].name;
    LibGetProgramName(name,mByte,sByte);
    if (name[0]==0||name[0]==0xFF)
    {
      if (mByte==6&&sByte==0)
        strcpy(name,"Spreadsheet");
      else
        strcpy(name,"AddIn");
    }
  }
}

/* Second level interface functions */

void clearDisp()
{
  LibClrDisp();
  LibPutDisp();
}

/* Interface handlers */

void activateEdit()
{
  createEvent(E_ACTIVATE,cList2[2].x+cList2[2].w-1,cList2[2].y+cList2[2].h-1,LPSW_NONE);
}

int onSave(Event* e, Control* c)
{
  if (e->type==E_PEN)
  {
    long size=FILE_BLOCK_SIZE*(long)LibGetFreeBlock();
    if (size<addins[listBox.selected].length)
    {
      dialogInfo(DICON_STOP,"Not enough space!");
      return 1;
    }

    strcpy(fileName,nameList[listBox.selected]);
    storeFlag=0;
    activateEdit();
    processForm(&inputForm);
    if (storeFlag)
    {
      char *str="Save this file:\r%?";
      char buffer[128];
      insertString(buffer,str,fileName);
      if (dialogSelect(DICON_EXCL,buffer))
      {
        /* Save selected addin here */
        createEvent(E_ACTIVATE,-1,-1,LPSW_NONE);
        pBar.size=addins[listBox.selected].length;
        processForm(&waitForm);
        if (saveFlag)
        {
           dialogInfo(DICON_OK,"AddIn saved!");
        }
        else
        {
           dialogInfo(DICON_STOP,"Saving failed!");
        }
        /* End of save */
        mainForm.status=FORM_INACTIVE;
      }
    }

    return 1;
  }
  return 0;
}

int onExit(Event* e, Control* c)
{
  if (e->type==E_PEN)
  {
    mainForm.status=FORM_INACTIVE;
    return 1;
  }
  return 0;
}

void mainHandler(Event* e, Form* f)
{
  if (e->type==E_ESC)
  {
    mainForm.status=FORM_INACTIVE;
  }
}

void checkName()
{
    if (fileName[0])
    {
      if (LibGetAllEntry(fileName, &mode, &submode))
      {
        dialogInfo(DICON_STOP,"File already exist!");
        activateEdit();
      }
      else
      {
        storeFlag=1;
        inputForm.status=FORM_INACTIVE;
      }
    }
    else
    {
      dialogInfo(DICON_STOP,"Invalid filename!");
      activateEdit();
    }
}

int onOk(Event* e, Control* c)
{
  if (e->type==E_PEN)
  {
    checkName();
    return 1;
  }
  return 0;
}

int onCancel(Event* e, Control* c)
{
  if (e->type==E_PEN)
  {
    inputForm.status=FORM_INACTIVE;
    return 1;
  }
  return 0;
}

void inputHandler(Event* e, Form* f)
{
  if (e->type==E_CANCEL)
  {
    inputForm.status=FORM_INACTIVE;
  }
  else if (e->type==E_COMPLETE)
  {
    checkName();
  }
}

void onActivateForm(Event* e, Form* f)
{
  if (e->type==E_ACTIVATE)
  {
    if (initFile(fileName)&&saveAddIn(listBox.selected))
    {
      LibWait(IB_500MWAIT);
      saveFlag=1;
    }
    waitForm.status=FORM_INACTIVE;
  }
}

void main()
{ 
  /* Detect all addins */
  clearDisp();
  LibWinIcnMsg(ICON_COFFEE,381,0x00);
  scanFlash();
  listBox.size=addCount;
  clearDisp();

  /* Init interface */
  LibTchInit();
  LibTchStackClr();

  initForm(&mainForm);
  initForm(&inputForm);
  initForm(&waitForm);

  /* Start interface */
  processForm(&mainForm);

  LibJumpMenu();
}
