/*
  ZProg PV Interface.
  (c) Alexey Makhmutov (aka Lerm) 2003,2004
*/

#include <ZiFace.h>
#include <stdrom.h>
#include <libc.h>
#include <l_define.h>

#include <DrawEx.h>

/* Service functions */
static int checkRect(int rx, int ry, int rw, int rh, int x, int y)
{
  return x>=rx&&x<(rx+rw)&&y>=ry&&y<(ry+rh);
}

static int checkControl(Control* control, int x, int y)
{
  return checkRect(control->x,control->y,control->w, control->h, x, y);
}

static int waitRelease(int x, int y, int w, int h)
{
  TCHSTS tch;
  do{
    LibTchWait(&tch);
  } while (tch.act != ACT_BREAK );
  return checkRect(x,y,w,h,tch.x,tch.y);
}

static int waitReleaseControl(Control* control)
{
  return waitRelease(control->x,control->y,control->w,control->h);
}

static void drawBox(int x, int y, int w, int h)
{
  LibGdsBox(x,y,x+w-1,y+h-1);
}

static void drawRoundBox(int x, int y, int w, int h)
{
  drawBox(x,y,w,h);
  LibGdsDotOff(x,y);
  LibGdsDotOff(x,y+h-1);
  LibGdsDotOff(x+w-1,y);
  LibGdsDotOff(x+w-1,y+h-1);
}

static void clearTchStack(int count)
{
  while(count-->0)
    LibTchStackPop();
}

/* Handlers */

/* Label */

#define LABEL_HEIGHT 9

static char* getString(char* point, int* len, int first)
{
  int l=*len;
  if (!first)
  {
    point+=l;
    if (*point=='\r')
      point++;
  }
  l=0;
  while(point[l]!=0&&point[l]!='\r')
    l++;
  *len=l;
  return point;
}

static int LibGetProStrSizeLim(byte type, byte *string, word lim)
{
  int c=0,w=0;
  while(c++<lim&&*string!=0)
    w+=LibGetProFontSize(type,*(string++));
  return w;
}

static void LibPutProStrExLim(byte type,int x,int y,byte *string,int x_size,word lim,byte style)
{
  int x_pos = x,c=0;
  while (*string&&c++<lim)
  {
    if ( x_pos + LibGetProFontSize(type, *string) > x + x_size )      
      break;    
    x_pos = LibPutProFontEx(type, *string, x_pos, y, style);    
    string++;
  }
}

static void labelCreator(Control* control)
{
  Label* label=getControl(control,Label);
  char* labelText=label->label;
  int y=control->y,h=0,len=0,maxH=160-LABEL_HEIGHT-y;
  
  labelText=getString(labelText,&len,TRUE);
  while(*labelText&&h<=maxH)
  {
    h+=LABEL_HEIGHT;
    labelText=getString(labelText,&len,FALSE);
  }
  control->h=h;
}

static void labelRender(Control* control)
{
  Label* label=getControl(control,Label);
  char* labelText=label->label;
  int x=control->x,y=control->y,w=control->w;
  byte font=(label->font)?(label->font):(IB_PFONT1);
  byte align=label->align;
  int len=0,textH=0,maxH=160-LABEL_HEIGHT-y;

  labelText=getString(labelText,&len,TRUE);
  while(*labelText&&textH<=maxH)
  {
    int textWidth=LibGetProStrSizeLim(font,labelText,len);
    int textX=x;
    if (align==ALIGN_RIGHT)
      textX=x+w-textWidth;
    else if (align==ALIGN_CENTER)
      textX=x+(w-textWidth)/2;
    LibPutProStrExLim(font,textX,y+textH,labelText,w,len,control->enabled?IB_GPFOVER:IB_GPFMESH);
    textH+=LABEL_HEIGHT;
    labelText=getString(labelText,&len,FALSE);
  }
}

/* Icon */
static void iconRender(Control* control)
{
  LibPutFarDataEx(control->x,control->y, getControl(control,Icon)->no, control->enabled?IB_FGPOVER:IB_FGPMESH);
}

/* Bitmap */

static void bitmapCreator(Control* control)
{
  byte far* image=getControl(control,Bitmap)->image;
  if (image)
  {
    int x=control->x, y=control->y;
    int w=*((word far*)image);
    int h=*((word far*)(image+2));

    if (x+w>160)
      w=160-x;
    if (y+h>160)
      h=160-y;

    control->w=w;
    control->h=h;
  }
}

static void bitmapRender(Control* control)
{
  byte far* image=getControl(control,Bitmap)->image;
  if (image)
  {
    LibPutGraphO(control->x,control->y,image,control->enabled?IB_GPOVER:IB_GPMESH);
  }
}

/* Buttons */

static void drawPushButton(int x, int y, int width, int height, char* label, byte state, byte enable)
{
  int strsize=LibGetProStrSize(IB_PFONT1,label);
  int tempX=x+width-3,tempY=y+height-3;
  int tempX2=tempX+2,tempY2=tempY+2;
  
  if (enable)
  {
    LibGdsLine(x+2,y,tempX,y);
    LibGdsLine(x,y+2,x,tempY);
    LibGdsLine(tempX2,y+2,tempX2,tempY);
    LibGdsLine(x+2,tempY2,tempX,tempY2);
  }
  else
  {
    LibGdsLineMesh(x+2,y,tempX,y);
    LibGdsLineMesh(x,y+2,x,tempY);
    LibGdsLineMesh(tempX2,y+2,tempX2,tempY);
    LibGdsLineMesh(x+2,tempY2,tempX,tempY2);
  }
  
  LibPutProStrEx(IB_PFONT1,x+((width-strsize)>>1),y+((height-8)>>1),label,strsize,enable?IB_GPFOVER:IB_GPFMESH);

  if (state)
    LibGdsReverse(x+1,y+1,tempX2-1,tempY2-1);

  LibGdsDotOn(x+1,y+1);
  LibGdsDotOn(tempX+1,y+1);
  LibGdsDotOn(x+1,tempY+1);
  LibGdsDotOn(tempX+1,tempY+1); 
}

static void drawCheckBox(int x, int y, int w, int h, char* label, byte state, byte enable)
{
  int strsize=LibGetProStrSize(IB_PFONT1,label);
  int state_pic=16+state;

  LibPutFarDataEx(x,y,state_pic,enable?IB_FGPOVER:IB_FGPMESH);
  LibPutProStrEx(IB_PFONT1,x+10,y,label,strsize,enable?IB_GPFOVER:IB_GPFMESH);
}

static void drawRadio(int x, int y, int w, int h, char* label, byte state, byte enable)
{
  int strsize=LibGetProStrSize(IB_PFONT1,label);
  int state_pic=13-state;

  LibPutFarDataEx(x,y,state_pic,enable?IB_FGPOVER:IB_FGPMESH);
  LibPutProStrEx(IB_PFONT1,x+10,y,label,strsize,enable?IB_GPFOVER:IB_GPFMESH);
}

typedef void (*BSubRender)(int x, int y, int w, int h, char* label, int state, int enabled);
static BSubRender bSubRenders[]=
                         {
                          (BSubRender)drawPushButton,
                          (BSubRender)drawCheckBox,
                          (BSubRender)drawRadio
                         };

static void buttonsRender(Control* control)
{
  Button* button=getControl(control,Button);

  bSubRenders[button->subType](control->x,control->y,
                               control->w,control->h,
                               button->label,button->state,control->enabled);
}

static void buttonsCreator(Control* control)
{
  Button* button=getControl(control,Button);
  int width=10;
  if (button->subType==B_CHECK||button->subType==B_RADIO)
  {
    if (button->label)   
      width+=LibGetProStrSize(IB_PFONT1,button->label);
    control->w=width;
    if (control->h<8)
      control->h=8;
  }
}

static int buttonsHandler(Event* event, Control* control)
{
  Button* button=getControl(control,Button);
  CHandler hand=control->custHand;
  int eType=event->type;

  if (eType==E_PEN)
  {
    int rez=1;
    int subType=button->subType;
    if (subType==B_PRESS)
    {
      button->state=TRUE;
      drawControl(control);
    }
    rez=waitReleaseControl(control);  
    if (subType==B_PRESS)
      button->state=FALSE;
    else if (subType==B_CHECK&&rez)
      button->state=!button->state;
    drawControl(control);
    if (rez&&hand)
      hand(event,control);
    return 1;
  }
  else if (hand&&(eType==E_LEVER||eType==E_ACTIVATE))
    return hand(event,control);

  return 0;
}

/* Radio-group handlers */
static void radioCreator(Control* control)
{
  RadioGroup* group=getControl(control,RadioGroup);
  Button** list=group->group;
  int size=group->size,w=0,h=0,i;
  for(i=0; i<size; i++)
  {
    if (list[i]->subType==B_RADIO)
    {
      int width=10;
      h+=10;
      if (list[i]->label)   
        width+=LibGetProStrSize(IB_PFONT1,list[i]->label);
      if (width>w)
        w=width;
    }
  }
  if (w>control->w)
    control->w=w;
  if (h>control->h)
    control->h=h;
}

static void radioRender(Control* control)
{
  RadioGroup* group=getControl(control,RadioGroup);
  Button** list=group->group;
  int size=group->size,x=control->x,y=control->y,w=control->w,h=control->h;
  int i;
  for(i=0; i<size; i++, y+=10)
    bSubRenders[list[i]->subType](x,y,w,h,list[i]->label,list[i]->state,control->enabled); 
}

static int radioHandler(Event* event, Control* control)
{
  RadioGroup* group=getControl(control,RadioGroup);
  Button** list=group->group;
  CHandler hand=control->custHand;

  int size=group->size,i;
  if (event->type==E_PEN)
  {
    int num=(event->y-control->y)/10;
    if (waitRelease(control->x,control->y+num*10,control->w,10))
    {
      for(i=0; i<size; i++)
        list[i]->state=(i==num);
      drawControl(control);
    }
    if (hand)
      hand(event,control);
    return 1;
  }
  return 0;
}

/* Listbox */

#define LIST_ITEM_HEIGHT  10
#define LIST_ITEM_MARGIN  2
#define LIST_ARROW_HEIGHT 10
#define LIST_SCROLL_WIDTH 11

static void arrow(int x, int y,  int h, int dx, int dy, int mesh)
{
  int divisor = (dy>=0 ? dy : -dy) + (dx>=0 ? dx : -dx);
  int incY1 = (dy+dx) / divisor;
  int incY2 = (dy-dx) / divisor;
  int incX1 = (dx-dy) / divisor;
  int incX2 = (dx+dy) / divisor;
  int x2 = x, y2 = y;
  while (--h >= 0)
  {
    if (mesh)
      LibGdsLineMesh(x, y, x2, y2);
    else
      LibGdsLine(x, y, x2, y2);
    x  += incX1;
    x2 += incX2;
    y  += incY1;
    y2 += incY2;
  }
}

static void listRender(Control* control)
{
  ListBox* listBox=getControl(control,ListBox);
  char** list=listBox->list;
  byte buffer[4+LIST_ITEM_HEIGHT*20]={160,0,LIST_ITEM_HEIGHT,0};
  int i;

  int x=control->x, y=control->y, w=control->w, h=control->h;
  int itemsCount=listBox->size, firstItem=listBox->first, selectedItem=listBox->selected;
  
  int itemsAreaX, itemsAreaY, itemsAreaWidth, itemsAreaHeight;
  int itemsActiveCount,itemsVisibleCount;

  int captureY=y+h-2;
  int captureHeight=LIST_ITEM_HEIGHT;
  
  itemsAreaX=x+LIST_ITEM_MARGIN;
  itemsAreaY=y+LIST_ITEM_MARGIN;
  itemsAreaWidth =w-2*LIST_ITEM_MARGIN;
  itemsAreaHeight=h-LIST_ITEM_MARGIN-1;

  itemsActiveCount =itemsAreaHeight/LIST_ITEM_HEIGHT;
  itemsVisibleCount=itemsActiveCount+1;

  drawRoundBox(x,y,w,h);

  if (itemsActiveCount<itemsCount)
  {
    int scrollHeight=h-2*LIST_ARROW_HEIGHT;
    int canUp=firstItem>0, canDown=(firstItem+itemsActiveCount)<itemsCount;
    int scrollX=x+w-LIST_SCROLL_WIDTH-1;
    int caretX=scrollX+1, caretY, caretWidth=LIST_SCROLL_WIDTH-2, caretHeight;
    int cX=scrollX+LIST_SCROLL_WIDTH/2; /* Center of arrow */
    
    itemsAreaWidth-=LIST_SCROLL_WIDTH+1;
    LibGdsLine(scrollX-1,y,scrollX-1,y+h-1);

    /* Draw arrows */
    arrow(cX, y+3,   5, 0,  1, !canUp);
    arrow(cX, y+h-4, 5, 0, -1, !canDown);

    /* Draw scroll area */
    LibReverse(caretX,y+LIST_ARROW_HEIGHT,caretWidth,scrollHeight);
    LibMesh(caretX,y+LIST_ARROW_HEIGHT,caretWidth,scrollHeight);

    /* Draw scroll caret */
    if (itemsActiveCount<itemsCount)
    {
      caretHeight=scrollHeight*itemsActiveCount/itemsCount;
      if (!canDown)
        caretY=y+h-LIST_ARROW_HEIGHT-caretHeight;
      else
        caretY=y+LIST_ARROW_HEIGHT+firstItem*scrollHeight/itemsCount;
    }
    else
    {
      caretHeight=scrollHeight;
      caretY=y+LIST_ARROW_HEIGHT;
    }

    LibClrBox(caretX,caretY,caretWidth,caretHeight);
    drawBox(caretX,caretY,caretWidth,caretHeight);
  }

  if (captureY+captureHeight>160)
    captureHeight=160-captureY;
  buffer[2]=(byte)captureHeight;

  if (captureHeight>0)
    LibGetGraph(0, captureY, 160, captureHeight, buffer+4);

  for(i=firstItem; i<itemsCount&&i<(firstItem+itemsVisibleCount); i++)
  {
    LibStringDsp(list[i], itemsAreaX, itemsAreaY, itemsAreaX+itemsAreaWidth-1, IB_PFONT1);
    if (i==selectedItem)
      LibReverse(itemsAreaX-1,itemsAreaY-1,itemsAreaWidth+2,LIST_ITEM_HEIGHT);
    itemsAreaY+=LIST_ITEM_HEIGHT;
  }
  if (captureHeight>0)
    LibPutGraph(0, captureY, buffer);
}

static int waitAction(int x, int y, int w, int h)
{
  TCHTBL tblAct[]={0,0,0,0,ACT_BREAK | ACT_BREAK_IN | ACT_500MSEC,0xD000,0,
                   0,0,0,0,ACT_NONE,OBJ_END,0};
  TCHSTS tch;
  int status=1;

  tblAct[0].x1=x;
  tblAct[0].y1=y;
  tblAct[0].x2=x+w-1; 
  tblAct[0].y2=y+h-1;

  LibReverse(x,y,w,h);
  LibPutDispBox(x,y,w,h);

  LibTchStackPush(0);
  LibTchStackPush(tblAct);

  do{
    LibTchWait(&tch);
  } while(!(tch.act & (ACT_BREAK | ACT_BREAK_IN | ACT_500MSEC)));

  if (tch.act==ACT_BREAK || tch.act==ACT_BREAK_IN)
    status=0;

  clearTchStack(2);

  if (!status)
  {
    LibReverse(x,y,w,h);
    LibPutDispBox(x,y,w,h); 
  }

  return status;
}

static int listHandler(Event* event, Control* control)
{
  ListBox* listBox=getControl(control,ListBox);

  int x=control->x,y=control->y,w=control->w,h=control->h;
  int itemsCount=listBox->size, firstItem=listBox->first, selectedItem=listBox->selected;
  int itemsAreaX, itemsAreaY, itemsAreaWidth, itemsAreaHeight;
  int itemsActiveCount,itemsVisibleCount;
  CHandler hand=control->custHand;

  itemsAreaX=x+LIST_ITEM_MARGIN;
  itemsAreaY=y+LIST_ITEM_MARGIN;
  itemsAreaWidth =w-2*LIST_ITEM_MARGIN;
  itemsAreaHeight=h-LIST_ITEM_MARGIN-1;

  itemsActiveCount =itemsAreaHeight/LIST_ITEM_HEIGHT;
  itemsVisibleCount=itemsActiveCount+1;

  if (event->type==E_PEN)
  {
    int eX=event->x,eY=event->y;
    int scrollX=x+w-LIST_SCROLL_WIDTH-1;

    if (itemsCount>itemsActiveCount&&eX>=scrollX)
    {
      int scrollHeight=h-2*LIST_ARROW_HEIGHT;
      int canUp=firstItem>0, canDown=(firstItem+itemsActiveCount)<itemsCount;
      int caretHeight=scrollHeight*itemsActiveCount/itemsCount, caretY;

      if (!canDown)
        caretY=y+h-LIST_ARROW_HEIGHT-caretHeight;
      else
        caretY=y+LIST_ARROW_HEIGHT+firstItem*scrollHeight/itemsCount;

      if (eY<y+LIST_ARROW_HEIGHT||eY>=y+h-LIST_ARROW_HEIGHT)
      {/* Up&Down */
        int dir=(eY<y+LIST_ARROW_HEIGHT)?-1:1;

        if (dir==-1&&canUp||dir==1&&canDown)
        {
          int status=1;
          firstItem+=dir;
          listBox->first=firstItem;
          drawControl(control);
          canUp=firstItem>0;
          canDown=(firstItem+itemsActiveCount)<itemsCount;

          while((dir==-1&&canUp||dir==1&&canDown)&&status)
          {
            status=waitAction(scrollX,dir==1?(y+h-LIST_ARROW_HEIGHT):y+1,LIST_SCROLL_WIDTH,LIST_ARROW_HEIGHT-1);
            if (status)
            {
              firstItem+=dir;
              listBox->first=firstItem;
              drawControl(control);
              canUp=firstItem>0;
              canDown=(firstItem+itemsActiveCount)<itemsCount;
            }
          }
        }
      }
      else 
      {/* Scroll area */
        TCHTBL tblAct[]={0,0,159,159,ACT_BREAK | ACT_BREAK_IN | ACT_MOVE | ACT_MOVE_IN,0xD000,0,
                         0,0,0,0,ACT_NONE,OBJ_END,0};
        TCHSTS tch;
        int relY0=-y-LIST_ARROW_HEIGHT-caretHeight/2;

        LibTchStackPush(0);
        LibTchStackPush(tblAct);
        do{
           int relY=relY0+eY;
           int newFirst;
           if (relY<0)
             newFirst=0;
           else
             newFirst=relY*itemsCount/scrollHeight;

           if (newFirst+itemsActiveCount>itemsCount)
             newFirst=itemsCount-itemsActiveCount;

           if (newFirst!=listBox->first)
           {
             listBox->first=newFirst;
             drawControl(control);
           }
           LibTchWait(&tch);
           eY=tch.y;
        } while (!(tch.act & (ACT_BREAK | ACT_BREAK_IN)));
        clearTchStack(2);
      }
    }
    else
    {
      int newSel=firstItem+(event->y-y)/LIST_ITEM_HEIGHT;
 
      if (newSel<itemsCount&&newSel!=listBox->selected)
      {
        listBox->selected=newSel;
        drawControl(control);
        waitReleaseControl(control);

        if (hand)
          hand(event,control);
      }
    }
    return 1;
  }
  else if (event->type==E_LEVER)
  {
    int lever=event->lever;

    if (lever & (LPSW_UP | LPSW_DOWN))
    {
      int callHand=0;

      if (lever==LPSW_UP&&selectedItem)
        selectedItem--;
      else if (lever==LPSW_DOWN&&(selectedItem+1)<itemsCount)
        selectedItem++;

      if (selectedItem<firstItem)
        firstItem=selectedItem;
      else if (selectedItem>=firstItem+itemsActiveCount)
        firstItem=selectedItem-itemsActiveCount+1;

      if (selectedItem!=listBox->selected)
        callHand=1;

      if (callHand||firstItem!=listBox->first)
      {
        listBox->selected=selectedItem;
        listBox->first=firstItem;
        drawControl(control);
      }

      if (callHand&&hand)
        hand(event,control);

      return 1;
    }
  }

  return 0;
}

/* Editbox */

#define EDIT_TEXT_MARGIN 2
#define EDIT_KB_Y        87
#define EDIT_KB_HEIGHT   73

static void editRender(Control* control)
{
  int x=control->x,y=control->y,w=control->w;
  int enabled=control->enabled;

  drawRoundBox(x,y,w,control->h);
  LibPutProStrEx(IB_PFONT1,x+EDIT_TEXT_MARGIN,y+EDIT_TEXT_MARGIN,
                 getControl(control,EditBox)->buffer,w-2*EDIT_TEXT_MARGIN,
                 enabled?IB_GPFOVER:IB_GPFMESH);
}

static int editHandler(Event* event, Control* control)
{
  EditBox* editBox=getControl(control,EditBox);
  if (event->type==E_PEN&&waitReleaseControl(control)||event->type==E_ACTIVATE)
  {
    int textX=control->x+EDIT_TEXT_MARGIN, textY=control->y+EDIT_TEXT_MARGIN;
    int textX2=textX+control->w-1-2*EDIT_TEXT_MARGIN;
    byte graphBuffer[EDIT_KB_HEIGHT*20+4]={160,0,EDIT_KB_HEIGHT,0};
    char* textBuffer=editBox->buffer;
    CHandler hand=control->custHand;

    TXTP inp;
    TCHTBL InpTch[3];
    T_SCR_POS scr;
    word blankGuide[]={0xFFFF}; /* No guide */
    TCHSTS tch;

    int process=TRUE; /* Work flag */

    /* Init now */
    inp.st_x=textX;
    inp.st_y=textY;
    inp.ed_x=textX2;
    inp.it_y=9;              /* Line height=9 */
    inp.MAXGYO=1;
    inp.font=IB_PFONT1;      /* Normal font */
    inp.csen=TRUE;
    inp.rtnen=2;             /* Only one line */
    inp.maxmj=editBox->maxNum;
    inp.txbf=textBuffer;
    inp.txtobj=OBJ_TXTTCH;
    inp.sbrobj=OBJ_SCR_BAR;
    inp.tchtb=InpTch;
    inp.s_pos=&scr;
    inp.gdcmt=blankGuide;
   
    LibGetGraph(0, EDIT_KB_Y, 160, EDIT_KB_HEIGHT, graphBuffer+4);
    LibGdsLine(0,EDIT_KB_Y,159,EDIT_KB_Y);
    LibTxtInit(&inp);
  
    /* Hook touch & lever events */
    LibTchStackPush(NULL);
    LibTchStackPush(TchAllDsp);
    LibTxtTchSet(&inp);
    LibTchStackPush(TchActionKey);  /* Hook lever (for left-right processing) */
    LibTchStackPush(TchHardIcon);   /* ! Process Hard-icons */

    /* Set cursor by touch */
    tch.obj=OBJ_TXTTCH;
    tch.act=ACT_MAKE;
    tch.x=event->x;
    tch.y=event->y;
    LibTxtInp(KEY_NONE,&tch,&inp);
    LibDispKey();
    LibPutDisp();

    /* Text loop */
    do{
      byte inpChar;     /* Readed character */
      if (LibTxtDsp(&inp))
        LibPutDisp();
  
      inpChar=LibGetKeyM(&tch);
  
      if (inpChar==KEY_RTN || inpChar==KEY_NEXT)
      {
        process=FALSE;/* Exit on Return or Next */
        createEvent(E_COMPLETE,-1,-1,LPSW_NONE);
      }
      else 
        if (inpChar==KEY_NONE && tch.obj!=OBJ_TXTTCH)
        {
          switch (tch.obj)
          {
            case OBJ_HIC_MBAR:
              {/* Context edit menu */
                word IdxEdt[]={20,21,22,PDNTBLEND};
                word IdxSys[]={PDNTBLEND}; 
                word IdxOpt[]={PDNTBLEND}; 
                word rez;
  
                LibPullDownInit(IdxEdt,IdxSys,IdxOpt);
                rez=LibPullDown();
                if (rez & PULLDOWN_EDIT)
                {
                  switch (rez & ~PULLDOWN_IDX_MASK)
                  {
                    case 1:
                      inp.txtst=TXTCUT;
                      break;
                    case 2:
                      inp.txtst=TXTCOPY;
                      break;
                    case 4:
                      inp.txtst=TXTPASTE;
                      break;
                  }
                }
              }
              break;
            case OBJ_LPSW_LEFT: /* Process left */
              inp.txtst=TXTCSUP;
              break;
            case OBJ_LPSW_RIGHT:/* Process right */
              inp.txtst=TXTCSDWN;
            case OBJ_LPSW_UP:
            case OBJ_LPSW_DOWN:
              break;
            case OBJ_HIC_ESC:
              process=FALSE;
              createEvent(E_CANCEL,-1,-1,LPSW_NONE);
              break;
            case OBJ_LPSW_PUSH:
              process=FALSE;
              createEvent(E_COMPLETE,-1,-1,LPSW_NONE);
              break;                
            default:
              if (tch.y<EDIT_KB_Y)  /* Exit if touch outside keyboard&editbox */
              {
                process=FALSE;
                if (tch.y!=-1&&tch.x!=-1)
                  createEvent(E_PEN,tch.x,tch.y,LPSW_NONE);
              }
          }
        }
  
      if (process)
        LibTxtInp(inpChar,&tch,&inp);
    } while (process);
  
    /* Clear touch stack */
    clearTchStack(5);
    LibPutGraph(0, EDIT_KB_Y, graphBuffer);
    LibPutDisp();
    drawControl(control);
    if (hand)
      hand(event,control);
    return 1;
  }

  return 0;
}

/* ProgressBar */

static void progressRender(Control* control)
{
  ProgressBar* progressBar=getControl(control,ProgressBar);
  unsigned long size=progressBar->size,current=progressBar->current;
  int x=control->x,y=control->y,w=control->w,h=control->h;
  int barX=x+2,barY=y+2;
  int barWidth=size?((w-4)*current/size):0;
  int barHeight=h-4;

  drawRoundBox(x,y,w,h);
  if (barWidth>0)
  {
    LibReverse(barX,barY,barWidth,barHeight);
    if (!control->enabled)
      LibMesh(barX,barY,barWidth,barHeight);
  }
}

/* Handlers tables */

typedef void (*CCreator)(Control*);
typedef void (*CRender)(Control*);

static CCreator creatorsList[]=
                {
                 (CCreator)labelCreator,     /* Label    */
                 NULL,                       /* Icon     */
                 (CCreator)bitmapCreator,    /* Bitmap   */
                 (CCreator)buttonsCreator,   /* Button   */
                 (CCreator)radioCreator,     /* Radio    */
                 NULL,                       /* List     */
                 NULL,                       /* Editbox  */
                 NULL                        /* Progress */
                };
static CRender  rendersList[] =
                {
                 (CRender)labelRender,       /* Label    */
                 (CRender)iconRender,        /* Icon     */
                 (CRender)bitmapRender,      /* Bitmap   */
                 (CRender)buttonsRender,     /* Button   */
                 (CRender)radioRender,       /* Radio    */
                 (CRender)listRender,        /* List     */
                 (CRender)editRender,        /* Editbox  */
                 (CRender)progressRender     /* Progress */
                };
static CHandler handlersList[]=
                {
                 NULL,                       /* Label    */
                 NULL,                       /* Icon     */
                 NULL,                       /* Bitmap   */
                 (CHandler)buttonsHandler,   /* Button   */
                 (CHandler)radioHandler,     /* Radio    */
                 (CHandler)listHandler,      /* List     */
                 (CHandler)editHandler,      /* Editbox  */
                 NULL                        /* Progress */
                };

/* Core functions */

void initControl(Control* control)
{
  int type=control->type;
  if (creatorsList[type])
    creatorsList[type](control);
}

void drawControl(Control* control)
{
  int type=control->type;

  if (rendersList[type]&&control->visible)
  {
    int x=control->x,y=control->y,w=control->w,h=control->h;
    LibClrBox(x,y,w,h);
    rendersList[type](control);
    LibPutDispBox(x,y,w,h);
  }
}

void initForm(Form* form)
{
  Control* list=form->list;
  int size=form->size,i;
  for(i=0; i<size; i++)
    initControl(list+i);
}

void drawForm(Form* form)
{
  Control* list=form->list;
  int size=form->size,i;
  for(i=0; i<size; i++)
    drawControl(list++);
}

#define EVENTS_MAXIMUM 16
static Event eventsStack[EVENTS_MAXIMUM];
static int eventsCount=0;

#define HOOKS_MAXIMUM  32
typedef struct
{
  int type;
  Control* control;
} EventHook;
static EventHook hooksStack[HOOKS_MAXIMUM];
static int hooksCount=0;

static Control* getHook(int type)
{
  if (hooksCount)
  {
    int i=hooksCount-1;
    while(hooksStack[i].control&&i>=0)
    {
      if (hooksStack[i].type==type)
        return hooksStack[i].control;
      i--;
    }
  }
  return NULL;
}

static int handleEvent(Event* event, Control* control)
{
  int cType=control->type;
  int eType=event->type;

  if (eType>=E_CUSTOM)
  {
    if (control->custHand)
      return control->custHand(event,control);
  }
  else
  {
    if (((eType!=E_PEN || eType==E_PEN && checkControl(control,event->x,event->y))) && 
         handlersList[cType] && control->enabled)
      return handlersList[cType](event,control);
  }
  return 0;
}

static void processEvent(Form* form)
{
  if (eventsCount)
  {
     Event event=eventsStack[--eventsCount];
     Control* control=getHook(event.type);

     if (control)
     {
       handleEvent(&event,control);
     }
     else
     {
       int size=form->size;
       int i;
       control=form->list;
       for(i=0; i<size; i++)
       {
         if (handleEvent(&event,control++))
           return;
       }
       if (form->hand)
         form->hand(&event,form);
     }
  }
}

#ifndef _HITACHI_
static int far decodeLever[]={LPSW_UP,LPSW_PUSH,LPSW_DOWN,LPSW_NONE,LPSW_NONE,LPSW_LEFT,LPSW_RIGHT};
#define MIN_LPSW OBJ_LPSW_UP
#else
static int far decodeLever[]={LPSW_PUSH,LPSW_UP,LPSW_DOWN,LPSW_LEFT,LPSW_RIGHT};
#define MIN_LPSW OBJ_LPSW_PUSH
#endif

void processForm(Form* form)
{
  TCHSTS tch;
  word obj;
  dword act;
  int x=form->x,y=form->y,w=form->w,h=form->h;
  int wasHook=0;

  LibTchStackPush(NULL);
  LibTchStackPush(TchHardIcon);
  LibTchStackPush(TchActionKey);
  LibTchStackPush(TchAllDsp);

  form->status=FORM_ACTIVE;
  LibOpenWindow(x,y,w,h); 
  LibPutDispBox(x,y,w,h);
  if (hooksCount)
  {
    createHook(E_PEN,NULL);
    wasHook=1;
  }

  drawForm(form);

  do {
    if (eventsCount)
      processEvent(form);
    else
    {
      LibTchWait(&tch);
      obj=tch.obj;
      act=tch.act;

      if (obj>=MIN_LPSW && obj<=OBJ_LPSW_RIGHT)
        createEvent(E_LEVER,-1,-1,decodeLever[obj-MIN_LPSW]);
      else if (obj==OBJ_HIC_ESC)
        createEvent(E_ESC,-1,-1,LPSW_NONE);
      else if (obj==OBJ_HIC_MBAR)
        createEvent(E_MENU,-1,-1,LPSW_NONE);
      else if (act & (ACT_MAKE | ACT_REPEAT))
        createEvent(E_PEN,tch.x,tch.y,LPSW_NONE);
    }
  } while(form->status==FORM_ACTIVE);
  LibCloseWindow();
  LibPutDispBox(x,y,w,h);
  clearTchStack(4);
  if (wasHook)
    removeHook(E_PEN,NULL);
}

int createEvent(int type, int x, int y, int lever)
{
  Event e;
  if (eventsCount>=EVENTS_MAXIMUM)
    return 0;

  e.type=type;
  e.x=x;
  e.y=y;
  e.lever=lever;
  eventsStack[eventsCount++]=e;
  return 1;
}

int createHook(int type, Control* control)
{
  EventHook h;
  if (hooksCount>=HOOKS_MAXIMUM)
    return 0;

  h.type=type;
  h.control=control;
  hooksStack[hooksCount++]=h;
  return 1;
}

int removeHook(int type, Control* control)
{
  if (hooksCount)
  {
    int i=hooksCount-1,j;
    do{
      if (hooksStack[i].type==type&&hooksStack[i].control==control)
      {
        for(j=i+1; j<hooksCount; j++)
          hooksStack[j-1]=hooksStack[j];
        hooksCount--;
        return 1;
      }
    }while(--i>=0);
  }
  return 0;
}